# coding: utf-8
from __future__ import unicode_literals

from .comments import API as CommentsAPI
from .endpoint import APIEndpoint
from .exceptions import BadQuery, HTTPError, UnavailableConvertor
from .models import Comment, Document
from .utils import version_lt
from .workflows import API as WorkflowsAPI

try:
    from typing import TYPE_CHECKING

    if TYPE_CHECKING:
        from typing import Any, Dict, List, Optional, Text, Union
        from .client import NuxeoClient
        from .models import Blob, Workflow
        from .operations import API as OperationsAPI
except ImportError:
    pass


class API(APIEndpoint):
    """ Endpoint for documents. """

    __slots__ = ("comments_api", "operations", "workflows_api")

    def __init__(
        self,
        client,  # type: NuxeoClient
        operations,  # type: OperationsAPI
        workflows,  # type: WorkflowsAPI
        comments,  # type: CommentsAPI
        endpoint=None,  # type: Text
        headers=None,  # type: Optional[Dict[Text, Text]]
    ):
        # type: (...) -> None
        self.operations = operations
        self.comments_api = comments
        self.workflows_api = workflows
        super(API, self).__init__(
            client, endpoint=endpoint, cls=Document, headers=headers
        )

    def get(self, uid=None, path=None):
        # type: (Optional[Text], Optional[Text]) -> Document
        """
        Get the detail of a document.

        :param uid: the uid of the document
        :param path: the path of the document
        :return: the document
        """
        return super(API, self).get(path=self._path(uid=uid, path=path))

    def post(self, document, parent_id=None, parent_path=None):
        # type: (Document, Optional[Text], Optional[Text]) -> Document
        """
        Create a document.

        :param document: the document to create
        :param parent_id: the id of the parent document
        :param parent_path: the path of the parent document
        :return: the created document
        """
        return super(API, self).post(
            document, path=self._path(uid=parent_id, path=parent_path)
        )

    create = post  # Alias for clarity

    def put(self, document):
        # type: (Document) -> Document
        """
        Update a document.

        :param document: the document to update
        :return: the updated document
        """
        return super(API, self).put(document, path=self._path(uid=document.uid))

    def delete(self, document_id):
        # type: (Text) -> None
        """
        Delete a document.

        :param document_id: the id of the document to delete
        """
        super(API, self).delete(self._path(uid=document_id))

    def exists(self, uid=None, path=None):
        # type: (Optional[Text], Optional[Text]) -> bool
        """
        Check if a document exists.

        :param uid: the id of the document to check
        :param path: the path of the document to check
        :return: True if it exists, else False
        """
        try:
            self.get(uid=uid, path=path)
            return True
        except HTTPError as e:
            if e.status != 404:
                raise e
        return False

    def add_permission(self, uid, params):
        # type: (Text, Dict[Text, Any]) -> None
        """
        Add a permission to a document.

        :param uid: the uid of the document
        :param params: the permissions to add
        """
        self.operations.execute(
            command="Document.AddPermission", input_obj=uid, params=params
        )

    def comment(self, uid, text):
        # type: (Text, Text) -> Comment
        """
        Add a new comment for a given document.

        :param uid: the ID of the document
        :param text: the content of the comment
        :return: the comment
        """
        return self.comments_api.create(uid, text)

    def comments(self, uid, params=None):
        # type: (Text, Any) -> List[Comment]
        """
        Get the comments of a document.

        Any additionnal arguments will be passed to the *params* parent's call.

        :param uid: the ID of the document
        :return: the list of comments
        """
        return self.comments_api.get(uid, params=params)

    def convert(self, uid, options):
        # type: (Text, Dict[Text, Text]) -> Union[Text, Dict[Text, Any]]
        """
        Convert a blob into another format.

        :param uid: the uid of the blob to be converted
        :param options: the target type, target format,
                        or converter for the blob
        :return: the response from the server
        """
        xpath = options.pop("xpath", "blobholder:0")
        adapter = "blob/{}/@convert".format(xpath)
        if (
            "converter" not in options
            and "type" not in options
            and "format" not in options
        ):
            raise BadQuery("One of (converter, type, format) is mandatory in options")

        try:
            return super(API, self).get(
                path=self._path(uid=uid), params=options, adapter=adapter, raw=True
            )
        except HTTPError as e:
            if "is not registered" in e.message:
                raise BadQuery(e.message)
            if (
                "is not available" in e.message
                or "UnsupportedOperationException" in e.message
            ):
                raise UnavailableConvertor(options)
            raise e

    def fetch_acls(self, uid):
        # type: (Text) -> Dict[Text, Any]
        """
        Fetch the ACLs of a document.

        :param uid: the uid of the document
        :return: the ACLs
        """
        req = super(API, self).get(
            path=self._path(uid=uid), cls=dict, headers=self.headers, enrichers=["acls"]
        )
        return req["contextParameters"]["acls"]

    def fetch_audit(self, uid):
        # type: (Text) -> Dict[Text, Any]
        """
        Fetch the audit of a document.

        :param uid: the uid of the document
        :return: the audit
        """
        return super(API, self).get(self._path(uid=uid), adapter="audit", cls=dict)

    def fetch_lock_status(self, uid):
        # type: (Text) -> Dict[Text, Any]
        """
        Fetch the lock status of a document.

        :param uid: the uid of the document
        :return: the lock status
        """
        headers = self.headers or {}
        headers.update({"fetch-document": "lock"})
        req = super(API, self).get(path=self._path(uid=uid), cls=dict, headers=headers)
        if "lockOwner" in req:
            return {"lockCreated": req["lockOwner"], "lockOwner": req["lockOwner"]}
        else:
            return {}

    def fetch_rendition(self, uid, name):
        # type: (Text, Text) -> Union[Text, bytes]
        """
        Fetch a rendition of a document.

        :param uid: the uid of the document
        :param name: the name of the rendition
        :return: the corresponding rendition
        """
        adapter = "rendition/{}".format(name)
        return super(API, self).get(path=self._path(uid=uid), raw=True, adapter=adapter)

    def fetch_renditions(self, uid):
        # type: (Text) -> List[Union[Text, bytes]]
        """
        Fetch all renditions of a document.

        :param uid: the uid of a document
        :return: the renditions
        """
        headers = self.headers or {}
        headers.update({"enrichers-document": "renditions"})

        req = super(API, self).get(path=self._path(uid=uid), cls=dict, headers=headers)
        return [rend["name"] for rend in req["contextParameters"]["renditions"]]

    def follow_transition(self, uid, name):
        # type: (Text, Text) -> Dict[Text, Any]
        """
        Follow a lifecycle transition.

        :param uid: the uid of the target document
        :param name: the name of the transition
        """
        params = {"value": name}
        return self.operations.execute(
            command="Document.FollowLifecycleTransition", input_obj=uid, params=params
        )

    def fetch_blob(self, uid=None, path=None, xpath="blobholder:0"):
        # type: (Optional[Text], Optional[Text], Text) -> Blob
        """
        Get the blob of a document.

        :param uid: the uid of the document
        :param path: the path of the document
        :param xpath: the xpath of the blob
        :return: the blob
        """
        adapter = "blob/{}".format(xpath)
        return super(API, self).get(
            path=self._path(uid=uid, path=path), raw=True, adapter=adapter
        )

    def get_children(self, uid=None, path=None, enrichers=None):
        # type: (Optional[Text], Optional[Text], Optional[List[str]]) -> List[Document]
        """
        Get the children of a document.

        :param uid: the uid of the document
        :param path: the path of the document
        :param enrichers: additionnal details to fetch at the same time, e.g.: ["permissions"]
        :return: the document children
        """
        return super(API, self).get(
            path=self._path(uid=uid, path=path), enrichers=enrichers, adapter="children"
        )

    def has_permission(self, uid, permission):
        # type: (Text, Text) -> bool
        """
        Check if a document has a permission.

        :param uid: the uid of the document
        :param permission: the permission to check
        :return: True if the document has it, False otherwise
        """
        req = super(API, self).get(
            path=self._path(uid=uid),
            cls=dict,
            headers=self.headers,
            enrichers=["permissions"],
        )
        return permission in req["contextParameters"]["permissions"]

    def lock(self, uid):
        # type: (Text) -> Dict[Text, Any]
        """ Lock a document. """
        return self.operations.execute(command="Document.Lock", input_obj=uid)

    def move(self, uid, dst, name=None):
        # type: (Text, Text, Optional[Text]) -> Dict[Text, Any]
        """
        Move a document and eventually rename it.

        :param uid: the uid of the target document
        :param dst: the destination
        :param name: the new name
        """
        params = {"target": dst}
        if name:
            params["name"] = name
        return self.operations.execute(
            command="Document.Move", input_obj=uid, params=params
        )

    def query(self, opts=None):
        # type: (Optional[Dict[Text, Text]]) -> Dict[Text, Any]
        """
        Run a query on the documents.

        :param opts: a query or a pageProvider
        :return: the corresponding documents
        """
        opts = opts or {}
        if "query" in opts:
            query = "NXQL"
        elif "pageProvider" in opts:
            query = opts["pageProvider"]
        else:
            raise BadQuery("Need either a pageProvider or a query")

        path = "query/{}".format(query)
        res = super(API, self).get(path=path, params=opts, cls=dict)
        res["entries"] = [
            Document.parse(entry, service=self) for entry in res["entries"]
        ]
        return res

    def remove_permission(self, uid, params):
        # type: (Text, Dict[Text, Text]) -> None
        """
        Remove a permission on a document.

        :param uid: the uid of the document
        :param params: the permission to remove
        """
        self.operations.execute(
            command="Document.RemovePermission", input_obj=uid, params=params
        )

    def trash(self, uid):
        # type: (Text) -> Dict[Text, Any]
        """
        Trash the document.

        :param uid: the uid of the document
        """
        if version_lt(self.client.server_version, "10.2"):
            input_obj = "doc:" + uid
            res_obj = self.operations.execute(
                command="Document.SetLifeCycle", input_obj=input_obj, value="delete"
            )
            res_obj["isTrashed"] = res_obj["state"] == "deleted"
            return res_obj

        return self.operations.execute(command="Document.Trash", input_obj=uid)

    def unlock(self, uid):
        # type: (Text) -> Dict[Text, Any]
        """ Unlock a document. """
        return self.operations.execute(command="Document.Unlock", input_obj=uid)

    def untrash(self, uid):
        # type: (Text) -> Dict[Text, Any]
        """
        Untrash the document.

        :param uid: the uid of the document
        """
        if version_lt(self.client.server_version, "10.2"):
            input_obj = "doc:" + uid
            res_obj = self.operations.execute(
                command="Document.SetLifeCycle", input_obj=input_obj, value="undelete"
            )
            res_obj["isTrashed"] = res_obj["state"] == "deleted"
            return res_obj

        return self.operations.execute(command="Document.Untrash", input_obj=uid)

    def workflows(self, document):
        # type: (Document) -> Union[Workflow, List[Workflow]]
        """ Get the workflows of a document. """
        path = "id/{}/@workflow".format(document.uid)
        return super(WorkflowsAPI, self.workflows_api).get(
            endpoint=self.endpoint, path=path
        )

    def _path(self, uid=None, path=None):
        # type: (Optional[Text], Optional[Text]) -> Text
        if uid:
            path = "repo/{}/id/{}".format(self.client.repository, uid)
        elif path:
            path = "repo/{}/path{}".format(self.client.repository, path)
        return path



# coding: utf-8
from __future__ import unicode_literals

from base64 import b64encode

from ..compat import get_bytes, get_text
from ..exceptions import NuxeoError
from ..utils import get_digest_hash

try:
    from typing import TYPE_CHECKING

    if TYPE_CHECKING:
        from typing import Text, Optional
except ImportError:
    pass


def make_portal_sso_token(timestamp, random, secret, username, digest_algorithm="md5"):
    # type: (int, Text, Text, Text, Optional[Text]) -> Text
    """Generate a token for SSO with Portals."""
    digester = get_digest_hash(digest_algorithm)
    if not digester:
        err = "Cannot compute token because of unknown digest algorithm: {!r}"
        raise NuxeoError(err.format(digest_algorithm))

    clear_token = ":".join([str(timestamp), random, secret, username])
    digester.update(get_bytes(clear_token))
    hashed_token = digester.digest()
    return get_text(b64encode(hashed_token))



# coding: utf-8
from __future__ import unicode_literals

from random import randint
from time import time

from ..compat import get_bytes
from .base import AuthBase
from .utils import make_portal_sso_token

try:
    from typing import TYPE_CHECKING

    if TYPE_CHECKING:
        from typing import Optional, Text
        from requests import Request
except ImportError:
    pass


class PortalSSOAuth(AuthBase):
    """Attaches required HTTP headers for SSO with Portals.
    See https://doc.nuxeo.com/nxdoc/using-sso-portals/ for details.
    """

    __slots__ = ("digest_algorithm", "secret", "username")

    NX_USER = get_bytes("NX_USER")
    NX_TOKEN = get_bytes("NX_TOKEN")
    NX_RD = get_bytes("NX_RD")
    NX_TS = get_bytes("NX_TS")

    def __init__(self, username, secret, digest_algorithm="md5"):
        # type: (Text, Text, Optional[Text]) -> None
        self.username = username
        self.secret = secret
        self.digest_algorithm = digest_algorithm.lower()

    def __eq__(self, other):
        # type: (object) -> bool
        return all(
            [
                self.username == getattr(other, "username", None),
                self.secret == getattr(other, "secret", None),
                self.digest_algorithm == getattr(other, "digest_algorithm", None),
            ]
        )

    def __ne__(self, other):
        # type: (object) -> bool
        return not self == other

    def __call__(self, r):
        # type: (Request) -> Request
        timestamp = int(time() * 1000)
        random = str(randint(0, timestamp))
        token = make_portal_sso_token(
            timestamp,
            random,
            self.secret,
            self.username,
            digest_algorithm=self.digest_algorithm,
        )

        r.headers[self.NX_TS] = str(timestamp)
        r.headers[self.NX_RD] = random
        r.headers[self.NX_TOKEN] = token
        r.headers[self.NX_USER] = self.username
        return r



# coding: utf-8
from __future__ import unicode_literals

from sys import platform

# OS
LINUX = platform.startswith("linux")
MAC = platform == "darwin"
WINDOWS = platform == "win32"

# Force parameters verification for all operations
CHECK_PARAMS = False

# Chunk size to download files
CHUNK_SIZE = 8192  # 8 KiB

# API paths
DEFAULT_API_PATH = "api/v1"
DEFAULT_URL = "http://localhost:8080/nuxeo/"

# Default value for the:
#   - 'X-Application-Name' HTTP header
#   - 'applicationName' URL parameter
DEFAULT_APP_NAME = "Python client"

# Name of the HTTP header for idempotent requests
IDEMPOTENCY_KEY = "Idempotency-Key"

# Maximum size to not overflow when logging raw content of a HTTP response
LOG_LIMIT_SIZE = 5 * 1024 * 1024  # 5 MiB

# Retries for each HTTP call on conection error
MAX_RETRY = 5

# Backoff factor between each retry
# Ex: with 0.2 then sleep() will sleep for [0.2s, 0.4s, 0.8s, ...] between retries
# Ex: with 1 then sleep() will sleep for [1s, 2s, 4s, ...] between retries
RETRY_BACKOFF_FACTOR = 1

# HTTP methods we want to handle in retries
RETRY_METHODS = frozenset(["GET", "POST", "PUT", "DELETE"])

# HTTP status code to handle for retries
# 425 Too Early
# 500 Internal Server Error
# 503 Service Unavailable
# 504 Gateway Timeout
RETRY_STATUS_CODES = [429, 500, 503, 504]

# TCP keep-alive values
# Amount of time in seconds between successive keep-alives sent to probe an unresponsive peer
TCP_KEEPINTVL = 60
# Maximum time in seconds to keep the connection idle before sending probes
TCP_KEEPIDLE = 60

# Connection and read timeout, in seconds
TIMEOUT_CONNECT = 10
TIMEOUT_READ = 60 * 10

# Size of chunks for the upload
UPLOAD_CHUNK_SIZE = 20 * 1024 * 1024  # 20 MiB

# Upload providers
UP_AMAZON_S3 = "s3"



# coding: utf-8
from __future__ import unicode_literals

try:
    from typing import TYPE_CHECKING

    if TYPE_CHECKING:
        from requests import Request
except ImportError:
    pass


class AuthBase(object):
    """Base class that all auth implementations derive from.
    Copied from requests.auth to add __slots__ capability.
    """

    __slots__ = ()

    def __call__(self, r):
        # type: (Request) -> Request
        raise NotImplementedError("Auth hooks must be callable.")



# coding: utf-8
from __future__ import unicode_literals

import hashlib
import logging
import mimetypes
import sys
from distutils.version import StrictVersion

from nuxeo.compat import lru_cache
from nuxeo.constants import UP_AMAZON_S3

try:
    from typing import TYPE_CHECKING

    if TYPE_CHECKING:
        from requests import Response
        from _hashlib import HASH
        from typing import Any, Dict, List, Optional, Text, Tuple, Union
except ImportError:
    pass


logger = logging.getLogger(__name__)
WIN32_PATCHED_MIME_TYPES = {
    "image/pjpeg": "image/jpeg",
    "image/x-png": "image/png",
    "image/bmp": "image/x-ms-bmp",
    "audio/x-mpg": "audio/mpeg",
    "video/x-mpeg2a": "video/mpeg",
    "application/x-javascript": "application/javascript",
    "application/x-msexcel": "application/vnd.ms-excel",
    "application/x-mspowerpoint": "application/vnd.ms-powerpoint",
    "application/x-mspowerpoint.12": "application/vnd.openxmlformats-officedocument.presentationml.presentation",
}


def chunk_partition(file_size, desired_chunk_size, handler=""):
    # type: (int, int, Optional[Text]) -> Tuple[int, int]
    """Determine the chunk count and chunk size from
    given *file_size* and *desired_chunk_size*.

    There may be boundaries depending of the given upload *handler*.

    :param file_size: the file size to upload (aka the blob)
    :param desired_chunk_size: the desired chunk size
    :param handler: the upload handler to use
    :return: a tuple of the chunk count and chunk size
    """

    chunk_size_min = 1
    chunk_count_max = None

    # Amazon S3 has some limitations
    if handler == UP_AMAZON_S3:
        chunk_size_min = 1024 * 1024 * 5  # 5 MiB
        chunk_count_max = 10000

    # Unsure to have the minimum valid chunk size
    chunk_size = max(chunk_size_min, desired_chunk_size)

    # Compute the number of chunks
    chunk_count = file_size // chunk_size + (file_size % chunk_size > 0)

    # If the chunks count exceeds the maximum chunks count ...
    if chunk_count_max and chunk_count > chunk_count_max:
        # ... In that case, the chunk count will define the chunk size
        return chunk_partition(
            file_size, file_size // (chunk_count_max - 1), handler=handler
        )

    return chunk_count, chunk_size


def cmp(a, b):
    # type: (Union[None, Text, StrictVersion], Union[None, Text, StrictVersion]) -> int
    """
    cmp() does not exist anymore in Python 3.
    Note: this function cannot be decorated with lru_cache() because when
    *a* or *b* is a *StrictVersion* object, it is not hashable.
    And callers are cached anyway.
    """
    if a is None:
        if b is None:
            return 0
        return -1
    if b is None:
        return 1
    return (a > b) - (a < b)


def get_digest_algorithm(digest):
    # type: (Text) -> Optional[Text]

    # Available algorithms
    digesters = {
        32: "md5",
        40: "sha1",
        56: "sha224",
        64: "sha256",
        96: "sha384",
        128: "sha512",
    }

    # Ensure the digest is hexadecimal
    try:
        int(digest, 16) >= 0
    except (TypeError, ValueError):
        return None

    return digesters.get(len(digest), None)


def get_digest_hash(algorithm):
    # type: (Text) -> Optional[HASH]

    # Retrieve the hashlib function for the given digest, None if not found
    func = getattr(hashlib, algorithm, None)

    return func() if func else None


def get_digester(digest):
    # type: (Text) -> Optional[HASH]
    """
    Get the digester corresponding to the given hash.

    To choose the digester used by the server, see
    https://doc.nuxeo.com/nxdoc/file-storage-configuration/#configuring-the-default-blobprovider

    :param digest: the hash
    :return: the digester function
    """

    algo = get_digest_algorithm(digest)
    func = get_digest_hash(algo) if algo else None

    if not func:
        logger.warning("No valid hash algorithm found for digest %r", digest)

    return func


def guess_mimetype(filename):
    # type: (Text) -> Text
    """ Guess the mimetype of a given file. """
    mime_type, _ = mimetypes.guess_type(filename)
    if mime_type:
        if sys.platform == "win32":
            # Patch bad Windows MIME types
            # See http://bugs.python.org/issue15207 and NXP-11660
            mime_type = WIN32_PATCHED_MIME_TYPES.get(mime_type, mime_type)
        return mime_type

    return "application/octet-stream"


def get_response_content(response, limit_size):
    # type: (Response, int) -> Text
    """Log a server response."""
    # Do not use response.text as it will load the chardet module and its
    # heavy encoding detection mecanism. The server will only return UTF-8.
    # See https://stackoverflow.com/a/24656254/1117028 and NXPY-100.
    try:
        content = response.content.decode("utf-8", errors="replace")
        content_size = len(content)
        if content_size > limit_size:
            content = content[:limit_size]
            content = "{} [...] ({:,} bytes skipped)".format(
                content, content_size - limit_size
            )
    except (MemoryError, OverflowError):
        # OverflowError: join() result is too long (in requests.models.content)
        # Still check for memory errors, this should never happen; or else,
        # it should be painless.
        headers = response.headers
        content_size = int(headers.get("content-length", 0))
        content = "<no enough memory ({:,} bytes)>".format(content_size)

    return content


def json_helper(obj):
    # type: (Any) -> Dict[Text, Any]
    return obj.to_json()


def log_chunk_details(chunk_count, chunk_size, uploaded_chunks, blob_size):
    # type: (int, int, List[int], int) -> None
    """Simple helper to log an chunked upload details about chunks data.

    :param chunk_count: the number of chunks
    :param chunk_size: the size of each chunks
    :param uploaded_chunks: the list of already uploaded chunks
    """
    if chunk_count <= 1 or logger.getEffectiveLevel() > logging.DEBUG:
        return

    uploaded_chunks_count = len(uploaded_chunks)
    uploaded_data_length = min(blob_size, chunk_size * uploaded_chunks_count)
    msg = (
        "Computed chunks count is {:,d}"
        "; chunks size is {:,d} bytes"
        "; uploaded chunks count is {:,d}"
        " => uploaded data so far is {:,d} bytes."
    )
    details = msg.format(
        chunk_count, chunk_size, uploaded_chunks_count, uploaded_data_length
    )
    logger.debug(details)


@lru_cache(maxsize=128)
def version_compare(x, y):
    # type: (Text, Text) -> int
    """
    Compare version numbers using the usual x.y.z pattern.
    For instance, will result in:
        - 5.9.3 > 5.9.2
        - 5.9.3 > 5.8
        - 5.8 > 5.6.0
        - 5.10 > 5.1.2
        - 1.3.0524 > 1.3.0424
        - 1.4 > 1.3.0524
        - ...
    Also handles snapshots and hotfixes:
        - 5.9.4-SNAPSHOT > 5.9.3-SNAPSHOT
        - 5.9.4-SNAPSHOT > 5.9.3
        - 5.9.4-SNAPSHOT < 5.9.4
        - 5.9.4-SNAPSHOT < 5.9.5
        - 5.8.0-HF15 > 5.8
        - 5.8.0-HF15 > 5.7.1-SNAPSHOT
        - 5.8.0-HF15 < 5.9.1
        - 5.8.0-HF15 > 5.8.0-HF14
        - 5.8.0-HF15 > 5.6.0-HF35
        - 5.8.0-HF15 < 5.10.0-HF01
        - 5.8.0-HF15-SNAPSHOT > 5.8
        - 5.8.0-HF15-SNAPSHOT > 5.8.0-HF14-SNAPSHOT
        - 5.8.0-HF15-SNAPSHOT > 5.8.0-HF14
        - 5.8.0-HF15-SNAPSHOT < 5.8.0-HF15
        - 5.8.0-HF15-SNAPSHOT < 5.8.0-HF16-SNAPSHOT
    """

    # Handle None values
    if not all((x, y)):
        return cmp(x, y)

    ret = (-1, 1)

    x_numbers = x.split(".")
    y_numbers = y.split(".")
    while x_numbers and y_numbers:
        x_part = x_numbers.pop(0)
        y_part = y_numbers.pop(0)

        # Handle hotfixes
        if "HF" in x_part:
            hf = x_part.replace("-HF", ".").split(".", 1)
            x_part = hf[0]
            x_numbers.append(hf[1])
        if "HF" in y_part:
            hf = y_part.replace("-HF", ".").split(".", 1)
            y_part = hf[0]
            y_numbers.append(hf[1])

        # Handle snapshots
        x_snapshot = "SNAPSHOT" in x_part
        y_snapshot = "SNAPSHOT" in y_part
        if not x_snapshot and y_snapshot:
            # y is snapshot, x is not
            x_number = int(x_part)
            y_number = int(y_part.replace("-SNAPSHOT", ""))
            return ret[y_number <= x_number]
        elif not y_snapshot and x_snapshot:
            # x is snapshot, y is not
            x_number = int(x_part.replace("-SNAPSHOT", ""))
            y_number = int(y_part)
            return ret[x_number > y_number]

        x_number = int(x_part.replace("-SNAPSHOT", ""))
        y_number = int(y_part.replace("-SNAPSHOT", ""))
        if x_number != y_number:
            return ret[x_number - y_number > 0]

    if x_numbers:
        return 1
    if y_numbers:
        return -1

    return 0


@lru_cache(maxsize=128)
def version_compare_client(x, y):
    # type: (Text, Text) -> int
    """ Try to compare SemVer and fallback to version_compare on error. """

    # Ignore date based versions, they will be treated as normal versions
    if x and "-I" in x:
        x = x.split("-")[0]
    if y and "-I" in y:
        y = y.split("-")[0]

    try:
        return cmp(StrictVersion(x), StrictVersion(y))
    except (AttributeError, ValueError):
        return version_compare(x, y)


@lru_cache(maxsize=128)
def version_le(x, y):
    # type: (Text, Text) -> bool
    """ x <= y """
    return version_compare_client(x, y) <= 0


@lru_cache(maxsize=128)
def version_lt(x, y):
    # type: (Text, Text) -> bool
    """ x < y """
    return version_compare_client(x, y) < 0



# coding: utf-8
from __future__ import unicode_literals

from ..compat import get_bytes
from .base import AuthBase

try:
    from typing import TYPE_CHECKING

    if TYPE_CHECKING:
        from typing import Text
        from requests import Request
except ImportError:
    pass


class TokenAuth(AuthBase):
    """ Attaches Nuxeo Token Authentication to the given Request object. """

    __slots__ = ("token",)

    HEADER_TOKEN = get_bytes("X-Authentication-Token")

    def __init__(self, token):
        # type: (Text) -> None
        self.token = token

    def __eq__(self, other):
        # type: (object) -> bool
        return self.token == getattr(other, "token", None)

    def __ne__(self, other):
        # type: (object) -> bool
        return not self == other

    def __call__(self, r):
        # type: (Request) -> Request
        r.headers[self.HEADER_TOKEN] = self.token
        return r



# coding: utf-8
"""
Inspired from https://github.com/finbourne/lusid-sdk-python/blob/ef029ec/sdk/lusid/tcp/tcp_keep_alive_probes.py
See NXPY-183 and linked tickets for details.
"""
from __future__ import unicode_literals

import socket

from requests.adapters import HTTPAdapter
from urllib3 import HTTPSConnectionPool, PoolManager

from ..constants import LINUX, MAC, TCP_KEEPINTVL, TCP_KEEPIDLE, WINDOWS


class TCPKeepAliveValidationMethods(object):
    """
    This class contains a single method whose sole purpose is to set up TCP keep-alive probes on the socket for a
    connection. This is necessary for long running requests which will be silently terminated by the AWS Network Load
    Balancer which kills a connection if it is idle for more then 350 seconds.
    """

    __slots__ = ()

    @staticmethod
    def adjust_connection_socket(conn):
        """Adjusts the socket settings so that the client sends a TCP keep-alive probe over the connection."""

        sock = conn.sock

        if LINUX:
            # https://tldp.org/HOWTO/html_single/TCP-Keepalive-HOWTO/#setsockopt
            sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPIDLE, TCP_KEEPIDLE)
            sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPINTVL, TCP_KEEPINTVL)
            # Number of times keep-alives are repeated before a close when there is no response
            #   Note 1: defaults to 10 on Windows, cannot be changed
            #   Note 2: defaults to 8 on macOS, cannot be changed since Catalina (sysctl net.inet.tcp.keepcnt)
            sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPCNT, 8)
        elif MAC:
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
            # 0x10 = TCP_KEEPALIVE (https://opensource.apple.com/source/xnu/xnu-6153.141.1/bsd/netinet/tcp.h.auto.html)
            sock.setsockopt(socket.IPPROTO_TCP, 0x10, TCP_KEEPINTVL)
        elif WINDOWS:
            # The Windows API requires milliseconds,
            # that's why the "* 1000" as TCP_* constants are expressed in seconds
            opt = (1, TCP_KEEPIDLE * 1000, TCP_KEEPINTVL * 1000)
            sock.ioctl(socket.SIO_KEEPALIVE_VALS, opt)


class TCPKeepAliveHTTPSConnectionPool(HTTPSConnectionPool):
    """
    This class overrides the _validate_conn method in the HTTPSConnectionPool class. This is the entry point to use
    for modifying the socket as it is called after the socket is created and before the request is made.
    """

    def _validate_conn(self, conn):
        """Called right before a request is made, after the socket is created."""
        super(TCPKeepAliveHTTPSConnectionPool, self)._validate_conn(conn)
        TCPKeepAliveValidationMethods.adjust_connection_socket(conn)


class TCPKeepAlivePoolManager(PoolManager):
    """
    This pool manager has only had the *pool_classes_by_scheme* variable changed.
    This now points at our custom connection pools rather than the default connection pools.
    """

    def __init__(self, num_pools=10, headers=None, **connection_pool_kw):
        super(TCPKeepAlivePoolManager, self).__init__(
            num_pools=num_pools, headers=headers, **connection_pool_kw
        )
        self.pool_classes_by_scheme = {
            "https": TCPKeepAliveHTTPSConnectionPool,
        }


class TCPKeepAliveHTTPSAdapter(HTTPAdapter):
    """"Transport adapter that allows us to set TCP keep-alive options."""

    def init_poolmanager(self, connections, maxsize, **pool_kwargs):
        """Override the default pool manager."""
        self.poolmanager = TCPKeepAlivePoolManager(
            num_pools=connections, maxsize=maxsize, **pool_kwargs
        )
